using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using NUnit.Framework;

#if NET_3_5
using System.Linq;
#else
using Smdn.Collections;
#endif
using Smdn.IO;
using Smdn.Net.Pop3.Client.Session;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopClientOperationsDownloadTests {
    private static void GetMessageBody(out string messageBody, out string messageByteStuffedBody, out int octets)
    {
      GetMessageBody(null, out messageBody, out messageByteStuffedBody, out octets);
    }

    private static void GetMessageBody(string subject, out string messageBody, out string messageByteStuffedBody, out int octets)
    {
      messageBody = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: " + (subject ?? string.Empty) +
@"

..byte-stuffed
+OK response style line
-ERR response style line2
.
end of message
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      messageByteStuffedBody = messageBody.Replace("\r\n.", "\r\n..");

      octets = NetworkTransferEncoding.Transfer8Bit.GetByteCount(messageBody);
    }

    private void TestMessage(Action<PopPseudoServer, PopClient, string> action)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        // STAT
        server.EnqueueResponse("+OK 3 200\r\n");
        // LIST
        server.EnqueueResponse("+OK 3 200\r\n");
        // RETR
        server.EnqueueResponse("+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        action(server, client, messageBody);

        // QUIT
        server.EnqueueResponse("+OK\r\n");

        client.Disconnect(true);

        Assert.AreEqual(server.DequeueRequest(), "QUIT\r\n");
      });
    }

    private void TestFirstMessage(Action<PopPseudoServer, PopClient, string> action)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        // STAT
        server.EnqueueResponse("+OK 5 1200\r\n");
        // LIST
        server.EnqueueResponse("+OK 1 200\r\n");
        // RETR
        server.EnqueueResponse("+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        action(server, client, messageBody);

        // QUIT
        server.EnqueueResponse("+OK\r\n");

        client.Disconnect(true);

        Assert.AreEqual(server.DequeueRequest(), "QUIT\r\n");
      });
    }

    private void TestLastMessage(Action<PopPseudoServer, PopClient, string> action)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        // STAT
        server.EnqueueResponse("+OK 5 1200\r\n");
        // LIST
        server.EnqueueResponse("+OK 5 200\r\n");
        // RETR
        server.EnqueueResponse("+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        action(server, client, messageBody);

        // QUIT
        server.EnqueueResponse("+OK\r\n");

        client.Disconnect(true);

        Assert.AreEqual(server.DequeueRequest(), "QUIT\r\n");
      });
    }

    private void TestNoMessagesExist(Action<PopClient> action)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // STAT
        server.EnqueueResponse("+OK 0 0\r\n");

        action(client);
      });
    }

    private void TestAllMessages(Action<PopPseudoServer, PopClient, string[]> action)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        const int messageCount = 5;
        var messageBodies = new string[messageCount];
        var byteStuffedMessageBodies = new string[messageCount];

        for (var i = 0; i < messageCount; i++) {
          int octets;
          string messageBody, byteStuffedMessageBody;

          GetMessageBody("#" + (i + 1).ToString(), out messageBody, out byteStuffedMessageBody, out octets);

          messageBodies[i] = messageBody;
          byteStuffedMessageBodies[i] = byteStuffedMessageBody;
        }
        
        // LIST
        var listResponse = "+OK\r\n";

        for (var i = 0; i < messageCount; i++) {
          listResponse += string.Format("{0} 200\r\n", i + 1);
        }

        listResponse += ".\r\n";

        server.EnqueueResponse(listResponse);

        // RETR
        for (var i = 0; i < messageCount; i++) {
          server.EnqueueResponse(string.Format("+OK\r\n{0}.\r\n", byteStuffedMessageBodies[i]));
        }

        action(server, client, messageBodies);

        // QUIT
        server.EnqueueResponse("+OK\r\n");
        
        client.Disconnect(true);
        
        Assert.AreEqual(server.DequeueRequest(), "QUIT\r\n");
      });
    }

    [Test]
    public void TestDownloadMessageAsText()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        Assert.AreEqual(expectedResult,
                        client.DownloadMessageAsText(3L, Encoding.UTF8));

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 3\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 3\r\n");
      });
    }

    [Test]
    public void TestDownloadFirstMessageAsText()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        Assert.AreEqual(expectedResult,
                        client.DownloadFirstMessageAsText(Encoding.UTF8));

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 1\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 1\r\n");
      });
    }

    [Test, ExpectedExceptionAttribute(typeof(ArgumentOutOfRangeException))]
    public void TestDownloadFirstMessageAsTextNoMessagesExist()
    {
      TestNoMessagesExist(delegate(PopClient client) {
        client.DownloadFirstMessageAsText(Encoding.UTF8);
      });
    }

    [Test]
    public void TestDownloadLastMessageAsText()
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        Assert.AreEqual(expectedResult,
                        client.DownloadLastMessageAsText(Encoding.UTF8));

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 5\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 5\r\n");
      });
    }

    [Test, ExpectedExceptionAttribute(typeof(ArgumentOutOfRangeException))]
    public void TestDownloadLastMessageAsTextNoMessagesExist()
    {
      TestNoMessagesExist(delegate(PopClient client) {
        client.DownloadLastMessageAsText(Encoding.UTF8);
      });
    }

    [Test]
    public void TestDownloadAllMessagesAsText()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        var downloadedMessages = client.DownloadAllMessagesAsText(Encoding.UTF8).ToArray();

        Assert.AreEqual(expectedResults.Length, downloadedMessages.Length, "downloaded message count");

        Assert.AreEqual(server.DequeueRequest(), "LIST\r\n");

        for (var i = 0; i < expectedResults.Length; i++) {
          Assert.AreEqual(expectedResults[i], downloadedMessages[i]);
          Assert.AreEqual(server.DequeueRequest(), string.Format("RETR {0}\r\n", i + 1));
        }
      });
    }

    [Test]
    public void TestDownloadAllMessagesAsTextNoMessagesExist()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // LIST
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");

        CollectionAssert.IsEmpty(client.DownloadAllMessagesAsText(Encoding.UTF8));
      });
    }

    [Test]
    public void TestDownloadMessageAsByteArray()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        CollectionAssert.AreEqual(Encoding.UTF8.GetBytes(expectedResult),
                                  client.DownloadMessageAsByteArray(3L));

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 3\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 3\r\n");
      });
    }

    [Test]
    public void TestDownloadFirstMessageAsByteArray()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        CollectionAssert.AreEqual(Encoding.UTF8.GetBytes(expectedResult),
                                  client.DownloadFirstMessageAsByteArray());

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 1\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 1\r\n");
      });
    }

    [Test]
    public void TestDownloadLastMessageAsByteArray()
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        CollectionAssert.AreEqual(Encoding.UTF8.GetBytes(expectedResult),
                                  client.DownloadLastMessageAsByteArray());

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 5\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 5\r\n");
      });
    }

    [Test]
    public void TestDownloadMessageAsStream()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        using (var stream = client.DownloadMessageAsStream(3L)) {
          FileAssert.AreEqual(new MemoryStream(Encoding.UTF8.GetBytes(expectedResult)),
                              stream);
        }

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 3\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 3\r\n");
      });
    }

    [Test]
    public void TestDownloadFirstMessageAsStream()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        using (var stream = client.DownloadFirstMessageAsStream()) {
          FileAssert.AreEqual(new MemoryStream(Encoding.UTF8.GetBytes(expectedResult)),
                              stream);
        }

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 1\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 1\r\n");
      });
    }

    [Test]
    public void TestDownloadLastMessageAsStream()
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        using (var stream = client.DownloadLastMessageAsStream()) {
          FileAssert.AreEqual(new MemoryStream(Encoding.UTF8.GetBytes(expectedResult)),
                              stream);
        }

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 5\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 5\r\n");
      });
    }

    [Test]
    public void TestDownloadMessageAsConverter()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = client.DownloadMessageAs(3L, delegate(Stream stream) {
          return stream.ReadToEnd().Length;
        });

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 3\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 3\r\n");
      });
    }

    [Test]
    public void TestDownloadFirstMessageAsConverter()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = client.DownloadFirstMessageAs(delegate(Stream stream) {
          return stream.ReadToEnd().Length;
        });

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 1\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 1\r\n");
      });
    }

    [Test]
    public void TestDownloadLastMessageAsConverter()
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = client.DownloadLastMessageAs(delegate(Stream stream) {
          return stream.ReadToEnd().Length;
        });

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 5\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 5\r\n");
      });
    }

    [Test]
    public void TestDownloadAllMessagesAsConverter()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        var downloadedResults = client.DownloadAllMessagesAs(delegate(Stream stream) {
          return stream.ReadToEnd().Length;
        }).ToArray();
        
        Assert.AreEqual(expectedResults.Length, downloadedResults.Length, "downloaded message count");
        
        Assert.AreEqual(server.DequeueRequest(), "LIST\r\n");
        
        for (var i = 0; i < expectedResults.Length; i++) {
          Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResults[i]), downloadedResults[i]);
          Assert.AreEqual(server.DequeueRequest(), string.Format("RETR {0}\r\n", i + 1));
        }
      });
    }

    [Test]
    public void TestDownloadMessageAsFunc()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = client.DownloadMessageAs(3L, delegate(Stream stream, string arg) {
          Assert.AreEqual(arg, "arg");

          return stream.ReadToEnd().Length;
        }, "arg");

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 3\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 3\r\n");
      });
    }

    [Test]
    public void TestDownloadFirstMessageAsFunc()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = client.DownloadFirstMessageAs(delegate(Stream stream, string arg) {
          Assert.AreEqual(arg, "arg");

          return stream.ReadToEnd().Length;
        }, "arg");

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 1\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 1\r\n");
      });
    }

    [Test]
    public void TestDownloadLastMessageAsFunc()
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = client.DownloadLastMessageAs(delegate(Stream stream, string arg) {
          Assert.AreEqual(arg, "arg");

          return stream.ReadToEnd().Length;
        }, "arg");

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 5\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 5\r\n");
      });
    }

    [Test]
    public void TestDownloadAllMessagesAsFunc()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        var downloadedResults = client.DownloadAllMessagesAs(delegate(Stream stream, string arg) {
          Assert.AreEqual(arg, "arg");

          return stream.ReadToEnd().Length;
        }, "arg").ToArray();
        
        Assert.AreEqual(expectedResults.Length, downloadedResults.Length, "downloaded message count");
        
        Assert.AreEqual(server.DequeueRequest(), "LIST\r\n");
        
        for (var i = 0; i < expectedResults.Length; i++) {
          Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResults[i]), downloadedResults[i]);
          Assert.AreEqual(server.DequeueRequest(), string.Format("RETR {0}\r\n", i + 1));
        }
      });
    }

    private void TestFile(Action<string, string> action)
    {
      const string expectedFile = "expected.eml";
      const string actualFile = "actual.eml";

      try {
        if (File.Exists(expectedFile))
          File.Delete(expectedFile);
        if (File.Exists(actualFile))
          File.Delete(actualFile);

        action(expectedFile, actualFile);

        FileAssert.AreEqual(expectedFile, actualFile);
      }
      finally {
        if (File.Exists(expectedFile))
          File.Delete(expectedFile);
        if (File.Exists(actualFile))
          File.Delete(actualFile);
      }
    }

    [Test]
    public void TestDownloadMessageToFile()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        TestFile(delegate(string expectedFilePath, string actualFilePath) {
          File.WriteAllText(expectedFilePath, expectedResult, new UTF8Encoding(false));

          var file = client.DownloadMessageToFile(3L, actualFilePath);

          Assert.IsTrue(file.Exists, "downloaded file exists");
          StringAssert.EndsWith(actualFilePath, file.FullName);
        });

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 3\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 3\r\n");
      });
    }

    [Test]
    public void TestDownloadFirstMessageToFile()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        TestFile(delegate(string expectedFilePath, string actualFilePath) {
          File.WriteAllText(expectedFilePath, expectedResult, new UTF8Encoding(false));

          var file = client.DownloadFirstMessageToFile(actualFilePath);

          Assert.IsTrue(file.Exists, "downloaded file exists");
          StringAssert.EndsWith(actualFilePath, file.FullName);
        });

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 1\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 1\r\n");
      });
    }

    [Test]
    public void TestDownloadLastMessageToFile()
    {
       TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        TestFile(delegate(string expectedFilePath, string actualFilePath) {
          File.WriteAllText(expectedFilePath, expectedResult, new UTF8Encoding(false));

          var file = client.DownloadLastMessageToFile(actualFilePath);

          Assert.IsTrue(file.Exists, "downloaded file exists");
          StringAssert.EndsWith(actualFilePath, file.FullName);
        });

        Assert.AreEqual(server.DequeueRequest(), "STAT\r\n");
        Assert.AreEqual(server.DequeueRequest(), "LIST 5\r\n");
        Assert.AreEqual(server.DequeueRequest(), "RETR 5\r\n");
      });
    }

    [Test]
    public void TestDownloadAllMessagesToDirectory()
    {
      const string dir = "downloaded";

      try {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);

        Directory.CreateDirectory(dir);

        TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
          var files = client.DownloadAllMessagesToDirectory(dir);

          Assert.AreEqual(expectedResults.Length, files.Length, "downloaded file count");
          Assert.AreEqual(server.DequeueRequest(), "LIST\r\n");

          for (var i = 0; i < expectedResults.Length; i++) {
            var expectedFileName = string.Format("{0}{1}{2}.eml", dir, Path.DirectorySeparatorChar, i + 1);

            Assert.IsTrue(files[i].Exists, "downloaded file exists #{0}", i);
            StringAssert.EndsWith(expectedFileName, files[i].FullName, "file name #{0}", i);

            Assert.AreEqual(expectedResults[i],
                            File.ReadAllText(files[i].FullName),
                            "downloaded content #{0}", i);

            Assert.AreEqual(server.DequeueRequest(), string.Format("RETR {0}\r\n", i + 1));
          }
        });
      }
      finally {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);
      }
    }

    [Test]
    public void TestDownloadAllMessagesToDirectoryTargetDirectoryNotExist()
    {
      const string dir = "downloaded";
      
      try {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);

        TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
          var files = client.DownloadAllMessagesToDirectory(dir);

          Assert.IsTrue(Directory.Exists(dir));
          Assert.AreEqual(expectedResults.Length, files.Length, "downloaded file count");
          Assert.AreEqual(server.DequeueRequest(), "LIST\r\n");

          for (var i = 0; i < expectedResults.Length; i++) {
            var expectedFileName = string.Format("{0}{1}{2}.eml", dir, Path.DirectorySeparatorChar, i + 1);

            Assert.IsTrue(files[i].Exists, "downloaded file exists #{0}", i);
            StringAssert.EndsWith(expectedFileName, files[i].FullName, "file name #{0}", i);

            Assert.AreEqual(expectedResults[i],
                            File.ReadAllText(files[i].FullName),
                            "downloaded content #{0}", i);

            Assert.AreEqual(server.DequeueRequest(), string.Format("RETR {0}\r\n", i + 1));
          }
        });
      }
      finally {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);
      }
    }

    [Test]
    public void TestDownloadAllMessagesToDirectoryArgumentNull()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        Assert.Throws<ArgumentNullException>(delegate {
          client.DownloadAllMessagesToDirectory(null);
        });
      });
    }

    [Test]
    public void TestDownloadAllMessagesToDirectoryArgumentEmpty()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        Assert.Throws<ArgumentException>(delegate {
          client.DownloadAllMessagesToDirectory(string.Empty);
        });
      });
    }

    [Test]
    public void TestDownloadAllMessagesToFiles()
    {
      const string dir = "downloaded";

      try {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);

        Directory.CreateDirectory(dir);

        TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
          var expectedFiles = new List<string>();

          var files = client.DownloadAllMessagesToFiles(delegate(PopMessageInfo msg) {
            var file = Path.Combine(dir, msg.MessageNumber.ToString("D4"));

            expectedFiles.Add(file);

            return file;
          });

          Assert.AreEqual(server.DequeueRequest(), "LIST\r\n");

          for (var i = 0; i < expectedResults.Length; i++) {
            Assert.IsTrue(files[i].Exists, "downloaded file exists #{0}", i);
            StringAssert.EndsWith(expectedFiles[i], files[i].FullName, "file name #{0}", i);

            Assert.AreEqual(expectedResults[i],
                            File.ReadAllText(files[i].FullName),
                            "downloaded content #{0}", i);

            Assert.AreEqual(server.DequeueRequest(), string.Format("RETR {0}\r\n", i + 1));
          }
        });
      }
      finally {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);
      }
    }

    [Test]
    public void TestDownloadAllMessagesToFilesTargetDirectoryNotExist()
    {
      const string dir = "downloaded";
      
      try {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);

        TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
          Assert.Throws<DirectoryNotFoundException>(delegate {
            client.DownloadAllMessagesToFiles(delegate(PopMessageInfo msg) {
              return Path.Combine(dir, msg.MessageNumber.ToString("D4"));
            });
          });

          Assert.AreEqual(server.DequeueRequest(), "LIST\r\n");
          Assert.AreEqual(server.DequeueRequest(), "RETR 1\r\n");
        });
      }
      finally {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);
      }
    }

    [Test]
    public void TestDownloadAllMessagesToFilesArgumentNull()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        Assert.Throws<ArgumentNullException>(delegate {
          client.DownloadAllMessagesToFiles(null);
        });
      });
    }
  }
}