using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using NUnit.Framework;

using Smdn.Net.Pop3.Client.Session;
using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Client;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopClientOperationsOneShotTests {
    [Test, ExpectedException(typeof(PopAuthenticationException))]
    public void TestInternalDownloadAppropriateAuthMechanismNotFound()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");

        PopClient.GetMessageCount(new Uri("pop://user@" + server.HostPort), "pass");
      }
    }

    [Test]
    public void TestInternalDownloadAuthenticateWithSpecifiedMechanism()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA
        server.EnqueueResponse("+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        // AUTH PLAIN
        server.EnqueueResponse("+OK\r\n");
        // STAT
        server.EnqueueResponse("+OK 5 1200\r\n");
        // QUIT
        server.EnqueueResponse("+OK\r\n");

        PopClient.GetMessageCount(new Uri("pop://user;AUTH=PLAIN@" + server.HostPort), "pass");

        StringAssert.StartsWith("CAPA", server.DequeueRequest());
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
      }
    }

    [Test]
#if DEBUG
    [Ignore("causes RemotingException")]
#endif
    public void TestInternalDownloadSecurePort()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA
        server.EnqueueResponse("+OK\r\n" +
                               "SASL STLS\r\n" +
                               ".\r\n");
        // STLS
        server.EnqueueResponse("-ERR\r\n");

        Assert.Throws<PopUpgradeConnectionException>(delegate {
          PopClient.GetMessageCount(new Uri("pops://user@" + server.HostPort), "pass");
        });
      }
    }

    private void TestOneShot(Action<PopPseudoServer> testAction)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA
        server.EnqueueResponse("+OK\r\n" +
                               "SASL CRAM-MD5\r\n" +
                               ".\r\n");
        // AUTH CRAM-MD5
        server.EnqueueResponse("+ PDQwMDEzNDQxMTIxNDM1OTQuMTI3MjQ5OTU1MEBsb2NhbGhvc3Q+\r\n");
        server.EnqueueResponse("+OK\r\n");

        testAction(server);
      }
    }

    [Test]
    public void TestGetMessageCount()
    {
      TestOneShot(delegate(PopPseudoServer server) {
        // STAT
        server.EnqueueResponse("+OK 5 1200\r\n");
        // QUIT
        server.EnqueueResponse("+OK\r\n");

        var ret = PopClient.GetMessageCount(new Uri("pop://user@" + server.HostPort), "pass");

        Assert.AreEqual(5, ret);

        server.DequeueRequest(); // CAPA
        server.DequeueRequest(); // AUTH
        server.DequeueRequest(); // AUTH response
        StringAssert.StartsWith("STAT", server.DequeueRequest());
      });
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadFirstMessageAsText(bool deleteAfterDownload)
    {
      TestOneShot(delegate(PopPseudoServer server) {
        // STAT
        server.EnqueueResponse("+OK 5 1200\r\n");
        // LIST
        server.EnqueueResponse("+OK 1 200\r\n");
        // RETR
        server.EnqueueResponse("+OK\r\n" +
                               "message#1\r\n" +
                               ".\r\n");
        
        if (deleteAfterDownload)
          // DELE
          server.EnqueueResponse("+OK\r\n");
        
        // QUIT
        server.EnqueueResponse("+OK\r\n");
        
        var ret = PopClient.DownloadFirstMessageAsText(new Uri("pop://user@" + server.HostPort),
                                                       "pass",
                                                       deleteAfterDownload,
                                                       Encoding.UTF8);
        
        Assert.AreEqual("message#1\r\n", ret);
        
        server.DequeueRequest(); // CAPA
        server.DequeueRequest(); // AUTH
        server.DequeueRequest(); // AUTH response
        StringAssert.StartsWith("STAT", server.DequeueRequest());
        StringAssert.StartsWith("LIST 1", server.DequeueRequest());
        StringAssert.StartsWith("RETR 1", server.DequeueRequest());
        
        if (deleteAfterDownload)
          StringAssert.StartsWith("DELE 1", server.DequeueRequest());
      });
    }
    
    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadLastMessageAsText(bool deleteAfterDownload)
    {
      TestOneShot(delegate(PopPseudoServer server) {
        // STAT
        server.EnqueueResponse("+OK 5 1200\r\n");
        // LIST
        server.EnqueueResponse("+OK 5 200\r\n");
        // RETR
        server.EnqueueResponse("+OK\r\n" +
                               "message#5\r\n" +
                               ".\r\n");
        
        if (deleteAfterDownload)
          // DELE
          server.EnqueueResponse("+OK\r\n");
        
        // QUIT
        server.EnqueueResponse("+OK\r\n");
        
        var ret = PopClient.DownloadLastMessageAsText(new Uri("pop://user@" + server.HostPort),
                                                      "pass",
                                                      deleteAfterDownload,
                                                      Encoding.UTF8);
        
        Assert.AreEqual("message#5\r\n", ret);
        
        server.DequeueRequest(); // CAPA
        server.DequeueRequest(); // AUTH
        server.DequeueRequest(); // AUTH response
        StringAssert.StartsWith("STAT", server.DequeueRequest());
        StringAssert.StartsWith("LIST 5", server.DequeueRequest());
        StringAssert.StartsWith("RETR 5", server.DequeueRequest());
        
        if (deleteAfterDownload)
          StringAssert.StartsWith("DELE 5", server.DequeueRequest());
      });
    }
    
    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadAllMessagesAsText(bool deleteAfterDownload)
    {
      TestOneShot(delegate(PopPseudoServer server) {
        // LIST
        server.EnqueueResponse("+OK\r\n" +
                               "1 120\r\n" +
                               "2 120\r\n" +
                               "3 120\r\n" +
                               ".\r\n");
        // RETR #1
        server.EnqueueResponse("+OK\r\n" +
                               "message#1\r\n" +
                               ".\r\n");
        if (deleteAfterDownload)
          // DELE #1
          server.EnqueueResponse("+OK\r\n");
        
        // RETR #2
        server.EnqueueResponse("+OK\r\n" +
                               "message#2\r\n" +
                               ".\r\n");
        if (deleteAfterDownload)
          // DELE #2
          server.EnqueueResponse("+OK\r\n");
        
        // RETR #3
        server.EnqueueResponse("+OK\r\n" +
                               "message#3\r\n" +
                               ".\r\n");
        if (deleteAfterDownload)
          // DELE #3
          server.EnqueueResponse("+OK\r\n");
        
        // QUIT
        server.EnqueueResponse("+OK\r\n");
        
        var ret = PopClient.DownloadAllMessagesAsText(new Uri("pop://user@" + server.HostPort),
                                                      "pass",
                                                      deleteAfterDownload,
                                                      Encoding.UTF8);
        
        Assert.IsNotEmpty(ret);
        Assert.AreEqual(3, ret.Length);
        Assert.AreEqual("message#1\r\n", ret[0]);
        Assert.AreEqual("message#2\r\n", ret[1]);
        Assert.AreEqual("message#3\r\n", ret[2]);
        
        server.DequeueRequest(); // CAPA
        server.DequeueRequest(); // AUTH
        server.DequeueRequest(); // AUTH response
        StringAssert.StartsWith("LIST", server.DequeueRequest());
        StringAssert.StartsWith("RETR 1", server.DequeueRequest());
        if (deleteAfterDownload)
          StringAssert.StartsWith("DELE 1", server.DequeueRequest());
        StringAssert.StartsWith("RETR 2", server.DequeueRequest());
        if (deleteAfterDownload)
          StringAssert.StartsWith("DELE 2", server.DequeueRequest());
        StringAssert.StartsWith("RETR 3", server.DequeueRequest());
        if (deleteAfterDownload)
          StringAssert.StartsWith("DELE 3", server.DequeueRequest());
      });
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadFirstMessageToFile(bool deleteAfterDownload)
    {
      const string dest = "download.eml";

      try {
        if (File.Exists(dest))
          File.Delete(dest);

        TestOneShot(delegate(PopPseudoServer server) {
          // STAT
          server.EnqueueResponse("+OK 5 1200\r\n");
          // LIST
          server.EnqueueResponse("+OK 1 200\r\n");
          // RETR
          server.EnqueueResponse("+OK\r\n" +
                                 "message#1\r\n" +
                                 ".\r\n");
          
          if (deleteAfterDownload)
            // DELE
            server.EnqueueResponse("+OK\r\n");
          
          // QUIT
          server.EnqueueResponse("+OK\r\n");
          
          var file = PopClient.DownloadFirstMessageToFile(new Uri("pop://user@" + server.HostPort),
                                                          "pass",
                                                          deleteAfterDownload,
                                                          dest);
          
          Assert.IsTrue(file.Exists, "downloaded file exists");
          StringAssert.EndsWith(dest, file.FullName, "downloaded file name");
          Assert.AreEqual("message#1\r\n", File.ReadAllText(file.FullName, Encoding.UTF8));
          
          server.DequeueRequest(); // CAPA
          server.DequeueRequest(); // AUTH
          server.DequeueRequest(); // AUTH response
          StringAssert.StartsWith("STAT", server.DequeueRequest());
          StringAssert.StartsWith("LIST 1", server.DequeueRequest());
          StringAssert.StartsWith("RETR 1", server.DequeueRequest());
          
          if (deleteAfterDownload)
            StringAssert.StartsWith("DELE 1", server.DequeueRequest());
        });
      }
      finally {
        if (File.Exists(dest))
          File.Delete(dest);
      }
    }
    
    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadLastMessageToFile(bool deleteAfterDownload)
    {
      const string dest = "download.eml";
      
      try {
        if (File.Exists(dest))
          File.Delete(dest);

        TestOneShot(delegate(PopPseudoServer server) {
          // STAT
          server.EnqueueResponse("+OK 5 1200\r\n");
          // LIST
          server.EnqueueResponse("+OK 5 200\r\n");
          // RETR
          server.EnqueueResponse("+OK\r\n" +
            "message#5\r\n" +
            ".\r\n");
        
          if (deleteAfterDownload)
          // DELE
            server.EnqueueResponse("+OK\r\n");
        
          // QUIT
          server.EnqueueResponse("+OK\r\n");
        
          var file = PopClient.DownloadLastMessageToFile(new Uri("pop://user@" + server.HostPort),
                                                         "pass",
                                                         deleteAfterDownload,
                                                         dest);
        
          Assert.IsTrue(file.Exists, "downloaded file exists");
          StringAssert.EndsWith(dest, file.FullName, "downloaded file name");
          Assert.AreEqual("message#5\r\n", File.ReadAllText(file.FullName, Encoding.UTF8));

          server.DequeueRequest(); // CAPA
          server.DequeueRequest(); // AUTH
          server.DequeueRequest(); // AUTH response
          StringAssert.StartsWith("STAT", server.DequeueRequest());
          StringAssert.StartsWith("LIST 5", server.DequeueRequest());
          StringAssert.StartsWith("RETR 5", server.DequeueRequest());
        
          if (deleteAfterDownload)
            StringAssert.StartsWith("DELE 5", server.DequeueRequest());
        });
      }
      finally {
        if (File.Exists(dest))
          File.Delete(dest);
      }
    }
    
    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadAllMessagesToDirectory(bool deleteAfterDownload)
    {
      const string destdir = "download";
      
      try {
        if (Directory.Exists(destdir))
          Directory.Delete(destdir, true);

        TestOneShot(delegate(PopPseudoServer server) {
          // LIST
          server.EnqueueResponse("+OK\r\n" +
            "1 120\r\n" +
            "2 120\r\n" +
            "3 120\r\n" +
            ".\r\n");
          // RETR #1
          server.EnqueueResponse("+OK\r\n" +
            "message#1\r\n" +
            ".\r\n");
          if (deleteAfterDownload)
          // DELE #1
            server.EnqueueResponse("+OK\r\n");
        
          // RETR #2
          server.EnqueueResponse("+OK\r\n" +
            "message#2\r\n" +
            ".\r\n");
          if (deleteAfterDownload)
          // DELE #2
            server.EnqueueResponse("+OK\r\n");
        
          // RETR #3
          server.EnqueueResponse("+OK\r\n" +
            "message#3\r\n" +
            ".\r\n");
          if (deleteAfterDownload)
          // DELE #3
            server.EnqueueResponse("+OK\r\n");
        
          // QUIT
          server.EnqueueResponse("+OK\r\n");
        
          var files = PopClient.DownloadAllMessagesToDirectory(new Uri("pop://user@" + server.HostPort),
                                                               "pass",
                                                               deleteAfterDownload,
                                                               destdir);
        
          Assert.IsTrue(Directory.Exists(destdir));

          Assert.IsTrue(files[0].Exists, "downloaded file exists #0");
          StringAssert.EndsWith(Path.Combine(destdir, "1.eml"), files[0].FullName, "downloaded file name #0");
          Assert.AreEqual("message#1\r\n", File.ReadAllText(files[0].FullName, Encoding.UTF8));

          Assert.IsTrue(files[1].Exists, "downloaded file exists #1");
          StringAssert.EndsWith(Path.Combine(destdir, "2.eml"), files[1].FullName, "downloaded file name #1");
          Assert.AreEqual("message#2\r\n", File.ReadAllText(files[1].FullName, Encoding.UTF8));

          Assert.IsTrue(files[2].Exists, "downloaded file exists #2");
          StringAssert.EndsWith(Path.Combine(destdir, "3.eml"), files[2].FullName, "downloaded file name #2");
          Assert.AreEqual("message#3\r\n", File.ReadAllText(files[2].FullName, Encoding.UTF8));

          server.DequeueRequest(); // CAPA
          server.DequeueRequest(); // AUTH
          server.DequeueRequest(); // AUTH response
          StringAssert.StartsWith("LIST", server.DequeueRequest());
          StringAssert.StartsWith("RETR 1", server.DequeueRequest());
          if (deleteAfterDownload)
            StringAssert.StartsWith("DELE 1", server.DequeueRequest());
          StringAssert.StartsWith("RETR 2", server.DequeueRequest());
          if (deleteAfterDownload)
            StringAssert.StartsWith("DELE 2", server.DequeueRequest());
          StringAssert.StartsWith("RETR 3", server.DequeueRequest());
          if (deleteAfterDownload)
            StringAssert.StartsWith("DELE 3", server.DequeueRequest());
        });
      }
      finally {
        if (Directory.Exists(destdir))
          Directory.Delete(destdir, true);
      }
    }
  }
}